"""Module with all the Time traking utils functions."""

import time
import json
from typing import Tuple, Optional, TextIO

from .print_utils import print_separator, print_warn

# =================================================
#                   HELPER FUNCTIONS
# =================================================
    
def parse_seconds_to_minutes(sec: float) -> str:
    """
    Parses a duration in seconds into a formatted string with hours, minutes, and seconds.

    Args:
        sec (float): Number of seconds.

    Returns:
        str: Formatted time string (e.g., '01 hrs, 05 mins, 30.1234 sec').
    """
    hours = int(sec // 3600)
    minutes = int((sec % 3600) // 60)
    seconds = int(sec % 60)
    decimals = int((sec % 1) * 10000)

    if hours > 0:
        return f"{hours:02} hrs, {minutes:02} mins, {seconds:02}.{decimals:04} sec"
    elif minutes > 0:
        return f"{minutes:02} mins, {seconds:02}.{decimals:04} sec"
    else:
        return f"{seconds:02}.{decimals:04} sec"


def print_time(sec: float, n_files: Optional[int] = None, space: bool = False, prefix: str = "", sufix: str = "", out_file: Optional[TextIO] = None) -> None:
    """Given a certain number of seconds, parse it to Formatted time string (e.g., '01 hrs, 05 mins, 30.1234 sec').
    If not enough seconds for hours, just '05 mins, 30.1234 sec'.
    If neither enogh seconds for minuts, just parse '30.1234 sec'.
    Optionally you can add a 'number of files' to get avg metrics as well as extra config for better printing. 
    Optionally you can pass a file to print everyting there.

    Args:
        sec (float): Number of seconds
        n_files (Optional[int], optional): Number of files to add an avg. Defaults to None.
        space (bool, optional): To add a space before the print. Defaults to False.
        prefix (str, optional): To add a prefix before the print. Defaults to "".
        sufix (str, optional): To add a sufix after the print. Defaults to "".
        out_file (Optional[TextIO], optional): To print the line somewere that's not the std bash (keed None for bash). Defaults to None.
    """
    if space:
        print("")
    
    if not prefix.endswith(" "):
        prefix = f"{prefix} "
    
    if n_files is not None:
        message = f"{prefix} - {n_files:4} files in: {parse_seconds_to_minutes(sec)}{sufix}.\n"
        message += f" -  Per document: {parse_seconds_to_minutes(sec / n_files)}"
    else:
        message = f"{prefix}Time: {parse_seconds_to_minutes(sec)}{sufix}."

    if out_file:
        print(message, file=out_file)
    else:
        print(message)
    
# =================================================
#                   TIME TRACKER
# =================================================
        
class TimeTracker:
    """
    Class for tracking the time of a process. It allows to track the time of different points in the process and save them in a json file.
    It also allows to track the time of different laps in the process.
    """
    def __init__(self, name: str, start_track_now: bool = False ):
        self.name = name
        self.hist: dict[str, Tuple[float, float]] = dict()
        self.started: bool = False
        self.last_time: float = -1.0
        
        self.lap_hist = dict()
        self.lap_runing: bool = False
        self.lap_number: int = 0
        
        if start_track_now:
            self.start(verbose=False)
            
        print_separator(f"⏳ TIME TRACKER '{name}' INITIALIZED{'. STARTING NOW' if start_track_now else ''}! ⏳")
    
    def start(self, verbose: bool = True, space: bool = True):
        """
        Starts traking the time
        """
        self.started = True
        self.track("START", verbose=verbose, space=space)
        
    
    def track(self, tag: str, verbose: bool = True, space: bool = False, mute_warning: bool = False) -> float:
        """
        Track the time of a certain point and add it a tag. Return time since las track
        """
        if not self.started and not mute_warning:
            print_warn("WARNING: Traking without startng, will call start.")
            self.start()
            
        t = time.time()
        diff = t - self.last_time if self.last_time > 0 else 0
        
        if self.lap_runing:
            if tag in self.lap_hist:
                tag = f"{tag}_{self.lap_number}"
            self.lap_hist[tag] = (t, diff)
        else:
            if tag in self.hist:
                tag = f"{tag}_"
            self.hist[tag] = (t, diff)
        
        if verbose: 
            print_tag = tag if not self.lap_runing else f"{tag} lap {self.lap_number}"
            print_time(diff, prefix=f"⏳ {print_tag}", sufix=" ⏳", space=space)
            
        self.last_time = t
        return diff
    
    # ============================================================================
    #                              LAPS MANAGEMENT
    # ============================================================================
    def start_lap(self, N: int = None, verbose: bool = False, mute_warning: bool = False) -> int:
        """Starts a new lap with its oun metrics and returns the number of the current started lap

        Args:
            N (int, optional): Total number of potential laps. Defaults to None.
            verbose (bool, optional): Print the number of the lap. Defaults to False.
            mute_warning (bool, optional): Show or not warnings. Defaults to False.

        Returns:
            int: Number of the current started lap.
        """
        self.lap_runing = True
        self.lap_number += 1
        if len(self.lap_hist) > 0 and not mute_warning:
            print_warn("WARNING: Starting lap without finishing previous. The records will be overritten.")
            
        t = time.time()    
        self.lap_hist["START_LAP"] = (t, 0)
        self.last_time = t
        
        if verbose:
            print(f"⏳ Starting lap num {self.lap_number}{f'/{N} ' if N is not None else ''}⏳!")
            
        return self.lap_number
        
    def finish_lap(self):
        """Finish lap and add the point trak to the list
        """
        self.lap_runing = False
        
        t = time.time()
        self.lap_hist["FINISH_LAP"] = (t, t-self.lap_hist["START_LAP"][0])
        
        # Update possible previous times
        for tag, (t, diff) in self.lap_hist.items():
            if tag in self.hist:
                _, prev_diff = self.hist[tag]
                self.hist[tag] = (t, prev_diff + diff)
            else:
                self.hist[tag] = (t, diff)
                
                
        self.lap_hist = dict()
        
    # ============================================================================
    #                              STIMATE TIME
    # ============================================================================
    def stimate_lap_time(self, N: int, mute_warning: bool = False):
        """Stimate the time to finish N laps after all the alread finished laps

        Args:
            N (int): Total number of laps (including those already done)
            mute_warning (bool, optional): Whether or not mute the waning about not running laps. Defaults to False.
        """
        if not self.lap_runing and not mute_warning:
            print_warn("WARNING: Stimating lap without starting it. Returning...")
            return
            
        t_f_end = time.time()
        eta = (N - self.lap_number) * (t_f_end - self.hist["START"][0]) / self.lap_number
        print_time(
            t_f_end - self.lap_hist["START_LAP"][0], 
            prefix="Total ", 
            sufix=f". ETA: {parse_seconds_to_minutes(eta)}"
        )
    
    # ============================================================================
    #                              METRICS MANAGEMENT
    # ============================================================================
        
    def get_metrics(self, n: int = None, initial_tag: str = "START") -> dict:
        """
        Return a dict with all the metrics with the form: tag: (time, diff) 
        Added Normalized if n of samples is passed with the form: tag: (time, diff, diff/n) 
        
        initial_tag change it in case it hasn't been set as 'START' for the first track
        """
        t = time.time()
        if len(self.hist) > 0: 
            if initial_tag not in self.hist:
                print_warn(f"WARNING: Passed initial tag '{initial_tag}' not found in history. Setting to first.")
                initial_tag = next(iter(self.hist)) # Getting the firts added tag
            self.hist["TOTAL"] = (t, t - self.hist[initial_tag][0])
            
        else:
            print_warn("WARNING: Getting metrics with 0 tracked points. This will return an empty dict.")
        
        if n is not None:
            res_hist =  {
                tag: (time, diff, diff/n) for tag, (time, diff) in self.hist.items()
            }
        else:
            res_hist = self.hist.copy()
        
        if "START_LAP" in res_hist:
            res_hist.pop("START_LAP")
        return res_hist
        
    def save_metric(self, save_path: str, n: int = None) -> dict:
        """Compute metrics, save them into a file and return them 

        Args:
            save_path (str): Save path for the metrics
            n (int, optional): 'Number of files' processed to get and avg. Defaults to None.

        Returns:
            dict: Computed metrics
        """
        metrics = self.get_metrics(n)
        
        with open(save_path, "w") as f:
            json.dump(metrics, f)
            
        return metrics
        
    def print_metrics(self, n: int = None, out_file: TextIO = None) -> dict:
        """Compute and print the metrics. Optionally into a file.

        Args:
            n (int, optional): 'Number of files' processed to get and avg. Defaults to None.
            out_file (TextIO, optional): File where printing should be done. Defaults to None.

        Returns:
            dict: Computed metrics
        """
        metrics = self.get_metrics(n)
        metrics.pop('START', None)
        print("")
        if n is not None:
            print(f"Processed {n} files in total\n", file=out_file)
        
        for tag, records in metrics.items():
            diff = records[1]
            
            print_time(diff, n_files=n, prefix=tag, out_file=out_file)
                    
        return metrics
        



from .file_utils import load_json, check_dirs_existance, make_dirs, save_json
from .print_utils import print_clear_bash, print_color, print_error, print_separator, print_status, print_warn
from .time_tracker import *


import os
import json
from typing import Any

from .print_utils import print_warn

def save_json(save_path: str, content: Any):
    """
    Saves a Python object as a JSON file.

    Args:
        save_path (str): Full path including file name to save the JSON.
        content (Any): The content to save (must be JSON-serializable).

    Returns:
        Any: The original content.
    """
    print(f"Saving output at {save_path}...")
    
    # Extra safety for empty paths
    dir_path = os.path.dirname(save_path)
    if dir_path:
        os.makedirs(dir_path, exist_ok=True)

    with open(save_path, "w", encoding="utf-8") as out_json:
        json.dump(content, out_json, indent=4)
    
    return content

def load_json(save_path: str) -> Any:
    """
    Loads JSON content from a file. Returns an empty dict if the file does not exist.

    Args:
        save_path (str): Full path including file name to read the JSON from.

    Returns:
        Any: The loaded content, or an empty dict if the file does not exist.
    """
    if not os.path.exists(save_path):
        print_warn(f"NO FILE AT {save_path}. Returning empty dict...")
        return dict()

    print(f"Loading output from {save_path}...")
    with open(save_path, "r", encoding="utf-8") as out_json:
        content = json.load(out_json)
    
    return content




def check_dirs_existance(directories: list[str]) -> None:
    """
    Checks if all specified directories exist.

    Args:
        directories (list[str]): List of directory paths to check.

    Raises:
        KeyError: If any of the directories do not exist.
    """
    missing = [d for d in directories if not os.path.exists(d)]
    if missing:
        raise KeyError(f"Some paths were not found: {missing}")


def make_dirs(directories: list[str]) -> None:
    """
    Creates the specified directories if they do not already exist.

    Args:
        directories (list[str]): List of directory paths to create.
    """
    for directory in directories:
        os.makedirs(directory, exist_ok=True)


from typing import Literal

# ==========================================================================================
#                                       GENERAL
# ==========================================================================================

separators = {
    "separator_short" : "_"*32,
    "separator_normal": "_"*64,
    "separator_long"  : "_"*128,
    "separator_super" : "="*128,
}

colors = {
    "red":    "\033[31m",
    "green":  "\033[32m",
    "yellow": "\033[33m",
    "blue":   "\033[34m",
    "white":  "\033[0m",
}

Colors = Literal["red", "green", "blue", "yellow", "white"]

def print_separator(text: str, sep_type: Literal["SHORT", "NORMAL", "LONG", "SUPER", "START"] = "NORMAL") -> None:
    """Prints a text with a line that separes the bash outputs. The size of this line is controled by sep_type

    Args:
        text (str): Text to print.
        sep_type (Literal['SHORT', 'NORMAL', 'LONG', 'SUPER'], optional): Type of the separation line. Defaults to "NORMAL".
    """
    if sep_type == "SHORT":
        sep = separators["separator_short"]
    elif sep_type == "NORMAL":
        sep = separators["separator_normal"]
    elif sep_type == "LONG":
        sep = separators["separator_long"]
    elif sep_type == "SUPER" or sep_type == "START":
        sep = separators["separator_super"]
    else:
        sep = separator_normal
        print_warn("WARNING: No separator with that label")
    
    if sep_type == "SUPER":
        print(sep)
        print(f"{text:^{len(sep)}}")
        print(sep + "\n")
    elif sep_type == "START":
        print_color(sep + "\n", color="blue")
        print_color(f"{text:^{len(sep)}}\n", color="blue")
        print_color(sep + "\n", color="blue")
    else:
        print(sep)
        print(f"{text:^{len(sep)}}\n")


def print_color(text: str, color: Colors = "white", print_text: bool = True) -> str:
    """Prints the text with a certain color

    Args:
        text (str): Text to print
        color (Literal['red', 'green', 'blue', 'white'], optional): Color to use. Defaults to "white".
        print_text bool: Whether or not to print the color text (if false it will return it)

    Return: 
        str: Text with colors
    """
    if color == "red":
        color = colors["color_red"]
    elif color == "green":
        color = colors["color_green"]
    elif color == "blue":
        color = colors["color_blue"]
    elif color == "yellow":
        color = colors["color_yellow"]
    else:
        color = colors["color_reset"]

    text: str = f"{color}{text}{colors["color_reset"]}"

    if print_text:
        print(f"{text}")

    return text


def print_warn(text: str, color: Colors = "yellow") -> str:
    """Adds the text between teh following emoji ⚠️...⚠️

    Args:
        text (str): Text to print in warn
        color (Colors, optional): Color of the warning text. Defaults to "yellow".

    Returns:
        str: Text with color and emojis
    """
    return print_color(f"⚠️{text}⚠️", color=color)

def print_error(text: str, color: Colors = "red") -> str:
    """Adds the text between teh following emoji ❌...❌

    Args:
        text (str): Text to print in warn
        color (Colors, optional): Color of the error text. Defaults to "red".

    Returns:
        str: Text with color and emojis
    """
    return print_color(f"❌{text}❌", color=color)


# ==========================================================================================
#                                    CLEAR LINES
# ==========================================================================================
def print_status(msg: str):
    """Prints a dynamic status message on the same terminal line.

    Useful for updating progress or status in-place (e.g. during loops),
    preventing multiple lines of output.

    Args:
        msg (str): Message to display.
    """
    clear_line = " " * 200  # assume max 200 chars per line
    print(f"\r{clear_line}\r{msg}", end="", flush=True)


def clear_bash(n_lines: int = 1) -> None:
    """Cleans the bash output by removing the last n lines.

    Args:
        n_lines (int, optional): Number of lines to remove. Defaults to 1.
    """
    print("\033[F\033[K"*n_lines, end="")  # Move cursor up one line and clear that line

def print_clear_bash(text: str, n_lines: int = 1) -> None:
    """Cleans the bash output by removing the last n lines.

    Args:
        n_lines (int, optional): Number of lines to remove. Defaults to 1.
    """
    clear_bash(n_lines)
    print(text)